
      include "flush_unit_module.inc"
      include "mpi_data_module.inc"

      module ata_definitions_module

************************************************************************
*
*     Name: ata_definitions_module
*
*     Usage: ata definitions.
*
*     use ata_definitions_module
*
*     Revision History:
*
*     25-Dec-03 -- SAF; initial construction of ata_definitions_module.
*
************************************************************************

      implicit none
      save

      integer (kind = 4), parameter :: NO = 0, YES = 1
      integer (kind = 4), parameter :: TST_BW = 0, TST_LAT = 1
      integer (kind = 4), parameter :: DEFAULT_TEST_TYPE = TST_BW
      integer (kind = 4), parameter :: DEFAULT_BANDWIDTH_BUFFER_SIZE = 100000 
      integer (kind = 4), parameter :: DEFAULT_LATENCY_BUFFER_SIZE   = 40

      end module ata_definitions_module

      program ata
***********************************************************************
*
*     Name: ata
*
*     Purpose: measure MPI performance using the MPI_Alltoall()
*              routine.
*
*     Usage
* 
*     ata -n num_times [-B|-L] [-b buff_size] [-h] [-a]
*
*        -n num_times   : number of times to repreat the ATA message
*                         passing sequence.
*        -B | -L        : Bandwidth or Latency test selection; default
*                         is the Bandwidth test.
*        -b buff_size   : buffer size in Bytes.
*        -h             : print usage line to <stdout>.
*        -a             : print out info line for each MPI process.
*
*     Algorithm:
*
*       interrogate command line options;
*
*       t1 = MPI_Wtime()
*       do i = 1, num_times 
*          call MPI_Alltoall (inbuf,  buff_size, ..., outbuf, ...)
*          call MPI_Alltoall (outbuf, buff_size, ..., inbuf, ...)
*       end do
*       t2 = MPI_Wtime()
*
*       if (test_type == latency)
*          latency = (1.0e+6 * (t2 - t1)) / (2 * num_times)
*       elseif (test_type == bandwidth)
*          tot_bytes = (2 * num_times) * (BUFF_KIND * num_procs * bufsize)
*          bandwidth = 1.0e-06 * tot_bytes / (t2 - t1)
*       endif
*
*       report results
*
*     Revision History:
*
*     13-Jun-02 -- SAF; add header documentation.
*     13-Jun-02 -- SAF; convert from f77 to f90.
*     13-Jun-02 -- SAF; add support for announce_flag and abort_flag.
*     24-Jun-02 -- SAF; replace nsteps with num_times.
*     22-Dec-03 -- SAF; use mpi_data_module
*     22-Dec-03 -- SAF; make message passing buffers run-time allocatable
*                       and provide support for a user specifiable buffer
*                       size.
*     25-Dec-03 -- SAF; add support for separate bandwidth and latency 
*                       tests.
*     01-Jan-04 -- SAF; correct latency calculation and restructure 
*                       test so that num_times is the number of times 
*                       that the double MPI_Alltoall() sequence is
*                       executed.
*     03-Jan-04 -- SAF; add -B|-L options to usage line.
*
***********************************************************************

      use ata_definitions_module
      use mpi_data_module
      
      implicit none

      character (len = 4), parameter :: VERSION = "2.1a"

      real (kind = 8) :: t0, t1, trans_time, bandwidth, latency
      integer (kind =4) :: i, j, num_times, num_ops
      real (kind = 4) :: total_bytes

      real (kind = BUFF_KIND), dimension (:), allocatable :: inbuf, outbuf
      integer (kind = 4) :: alloc_status

      integer (kind = 4) :: announce_flag, abort_flag, buffer_size, test_type

*     Initialize the MPI library and get operating parameters.

      call MPI_Begin 

*     Write program header.      

      if (my_rank .eq. MASTER_RANK) then
         write (6, 100) version
         write (6, 120) BUFF_KIND
      endif

*     If this is the master, interrogate the command line options.

      if (my_rank .eq. MASTER_RANK) then
         call get_command_line_options (num_times, announce_flag, 
     &                                  abort_flag, buffer_size,
     &                                  test_type)
      endif

*     Broadcast the abort flag and if some error occurred in the
*     processing of the command line options, then terminate the
*     program.

      call MPI_Bcast (abort_flag, 1, MPI_INTEGER, 0,
     &                MPI_COMM_WORLD, mpi_ierr)
      if (abort_flag .eq. YES) then
         if (my_rank .eq. MASTER_RANK) then
             call print_usage
         endif
         call MPI_End
         call exit (0)
      endif

*     Broadcast command line options to all other MPI processes.

      call MPI_Bcast (num_times,     1, MPI_INTEGER, 0, 
     &                MPI_COMM_WORLD, mpi_ierr)
      call MPI_Bcast (buffer_size,   1, MPI_INTEGER, 0, 
     &                MPI_COMM_WORLD, mpi_ierr)
      call MPI_Bcast (test_type,     1, MPI_INTEGER, 0,
     &                MPI_COMM_WORLD, mpi_ierr)
      call MPI_Bcast (announce_flag, 1, MPI_INTEGER, 0, 
     &                MPI_COMM_WORLD, mpi_ierr)

*     If the announce flag is set, then call announce().

      if (announce_flag .eq. YES) then
         call announce 
      endif

*     Allocate and initialize the message buffer.

      allocate (inbuf(num_procs * buffer_size), stat = alloc_status)
      if (alloc_status .ne. 0) then
         write (6, *) "ata: input message buffer allocation error",
     &                ", alloc_status = ", alloc_status
         call exit (1)
      endif

      allocate (outbuf(num_procs * buffer_size), stat = alloc_status)
      if (alloc_status .ne. 0) then
         write (6, *) "ata: output message buffer allocation error",
     &                ", alloc_status = ", alloc_status
         call exit (1)
      endif

*     Initialize the message buffers.

      do i = 1, num_procs * buffer_size
         inbuf(i)  = float(my_rank)
         outbuf(i) = -float(my_rank)
      end do

*     Synchronize all processes.

      call MPI_Barrier (MPI_COMM_WORLD, mpi_ierr)

*     Perform the all-to-all transfers.

      if (my_rank .eq. MASTER_RANK) then
         t0 = MPI_Wtime()
      endif
      do i = 1, num_times

         call MPI_Alltoall (inbuf,  buffer_size, MPI_FLPT,
     &                      outbuf, buffer_size, MPI_FLPT,
     &                      MPI_COMM_WORLD, mpi_ierr)

         call MPI_Alltoall (outbuf, buffer_size, MPI_FLPT,
     &                      inbuf,  buffer_size, MPI_FLPT,
     &                      MPI_COMM_WORLD, mpi_ierr)

      end do
      if (my_rank .eq. MASTER_RANK) then
         t1 = MPI_Wtime()
      endif

*     If this is the master, compute and print out the results.

      if (my_rank .eq. MASTER_RANK) then

         num_ops     = 2 * num_times
         trans_time  = t1 - t0         
         total_bytes = real(BUFF_KIND)  * real (num_ops) * 
     &                 real (num_procs) * real (buffer_size)

         write (6, 110) trans_time
         write (6, 111) num_ops
         write (6, 112) real (num_ops) / trans_time
         write (6, 115) BUFF_KIND * buffer_size
         write (6, 113) total_bytes * 1.0e-06
         if (test_type .eq. TST_LAT) then
            latency = (1.0e+6 *trans_time) / real (num_ops)
            write (6, 116) latency
         endif
         if (test_type .eq. TST_BW) then
            bandwidth = (1.0e-6 * total_bytes) / trans_time
            write (6, 114) bandwidth
         endif

      endif

*     Close up shop MPI-wise.

      call MPI_Finalize (mpi_ierr)

*     Say bye-bye.

      call exit (0)

 100  format (' -- ****** ATA -- All-To-All Bandwidth and Latency Test,',
     &        ' version = ', a, ' ****** --')
 110  format(' Total time:                    ', f12.3, ' sec.')
 111  format(' Number of alltoall operations: ', i12)
 112  format(' Operation rate:                ', f12.1, 
     &       ' alltoall-ops/sec.')
 113  format(' Total number of bytes moved:   ', f12.3,
     &       ' 10e+06 bytes/proc')
 114  format(' Bandwidth:                     ', f12.3, 
     &       ' 10e+06 bytes/sec/proc')
 115  format(' Buffer size:                   ', i12, 
     &       ' bytes')
 116  format(' Latency:                       ', f12.2, 
     &       ' microsec')

 120  format (' Message passing data size: ', 14x, i2, ' bytes.')

      end program ata

      subroutine get_command_line_options (num_times, announce_flag, 
     &                                     abort_flag, buffer_size,
     &                                     test_type)
************************************************************************
*
*     Name: get_command_line_options
*
*     Purpose: get command line options.
*
*     Revision History:
*
*     13-Jun-02 -- SAF; add header documentation.
*     13-Jun-02 -- SAF; add support for announce_flag and abort_flag.
*     17-Jun-02 -- SAF; added code to better handle bad or missing 
*                       arguments for the '-n' option.
*     22-Dec-03 -- SAF; use mpi_data_module
*     22-Dec-03 -- SAF; rename get_command_line_options and add support
*                       for user specifiable buffer size.
*     25-Dec-03 -- SAF; add support for test type option.
*
************************************************************************

      use mpi_data_module
      use ata_definitions_module

      implicit none

      integer (kind = 4) :: num_times, abort_flag, announce_flag, buffer_size,
     &                      test_type

      character (len = 120) :: arg
      integer (kind = 4) :: num_args, iargc, inx_arg
      integer (kind = 4) :: num_times_found, buffer_size_found, test_type_found

*     Begin get_command_line_options.

*     Scan through the arguments.

      num_times_found   = NO
      buffer_size_found = NO
      abort_flag        = NO
      announce_flag     = NO
      test_type         = DEFAULT_TEST_TYPE

      num_args          = iargc()
      do inx_arg = 1, num_args     

         call getarg (inx_arg, arg)

         if (arg .eq. '-h') then
            abort_flag = YES
            return
         endif

         if (arg .eq. '-a') then
            announce_flag = YES
         endif

         if (arg .eq. '-B') then
            test_type = TST_BW
         endif

         if (arg .eq. '-L') then
            test_type = TST_LAT
         endif

         if (arg .eq. '-n') then

            num_times_found = YES
            if ((inx_arg + 1) .le. num_args) then

               call getarg (inx_arg + 1, arg)

               read (arg, *, err = 110) num_times
               go to 111
 110           write (6, *) "ata: undecipherable argument for the -n option"
               abort_flag = YES
               return
 111           continue
               
            else

               write (6, *) "ata: missing argument for the -n option"
               abort_flag = YES
               return

            endif
         endif

         if (arg .eq. '-b') then

            buffer_size_found = YES
            if ((inx_arg + 1) .le. num_args) then

               call getarg (inx_arg + 1, arg)

               read (arg, *, err = 120) buffer_size
               buffer_size = buffer_size / BUFF_KIND
               go to 121

 120           write (6, *) "ata: undecipherable argument for the -b option"
               abort_flag = YES
               return

 121           continue
               
            else

               write (6, *) "ata: missing argument for the -b option"
               abort_flag = YES
               return

            endif
         endif

      enddo

*     Sanity checks on the value of num_times.

      if (num_times_found .eq. NO) then
         write (6, *) "ata: missing required argument -n"
         abort_flag = YES
      elseif (num_times .le. 0) then
         write (6, *) "ata: illegal value for argument to the -n option"
         abort_flag = YES
      endif

*     If there was no buffer size option on the command line, then 
*     set the buffer size depending on the test type.

      if (buffer_size_found .eq. NO) then
         if (test_type .eq. TST_BW) then
            buffer_size = DEFAULT_BANDWIDTH_BUFFER_SIZE / BUFF_KIND
         endif
         if (test_type .eq. TST_LAT) then
            buffer_size = DEFAULT_LATENCY_BUFFER_SIZE / BUFF_KIND
         endif
      endif

*     Sanity checks on the value of buffer_size

      if (buffer_size .le. 0) then
         write (6, *) "ata: illegal value for argument to the -b option"
         abort_flag = YES
      endif

*     End get_command_line_options.

      return
      end subroutine get_command_line_options

      subroutine print_usage
************************************************************************
*
*     Name: print_usage
*
*     Purpose: print usage message to <stdout>.
*
*     Revision History:
*
*     13-Jun-02 -- SAF; add header documentation.
*     22-Dec-03 -- SAF; add -b argument to usage line.
*     03-Jan-04 -- SAF; add -B|-L options to usage line.
*
************************************************************************

*     Begin print_usage.

      write (6, *) "Usage: ata -n num_times [-B|-L] [-b buff_size] [-h] [-a]"

*     End print_usage.

      return
      end subroutine print_usage

