**#define VALIDATE

      include "flush_unit_module.inc"
      include "mpi_data_module.inc"

      module lbw_definitions_module

************************************************************************
*
*     Name: lbw_definitions_module
*
*     Usage: lbw definitions.
*
*     use lbw_definitions_module
*
*     Revision History:
*
*     23-Dec-03 -- SAF; initial construction of lbw_definitions_module.
*
************************************************************************

      implicit none
      save

      integer (kind = 4), parameter :: NO = 0, YES = 1
      integer (kind = 4), parameter :: TST_BW = 0, TST_LAT = 1
      integer (kind = 4), parameter :: DEFAULT_TEST_TYPE = TST_BW
      integer (kind = 4), parameter :: STYLE_SYNC = 0, STYLE_ASYNC = 1
      integer (kind = 4), parameter :: DEFAULT_BANDWIDTH_BUFFER_SIZE = 1000000
      integer (kind = 4), parameter :: DEFAULT_LATENCY_BUFFER_SIZE   = 40

      end module lbw_definitions_module

      program lbw
************************************************************************
*
*     Name: lbw
*
*     Purpose: measure MPI point-to-point bandwidth and latency.
*
*     Usage:
*
*     lbw -n num_times [-L|-B] [-b buff_size] [-s sync|async] [-h] [-a]
*
*        -L | -B        : optional latency (-L) or bandwidth (-B) test type;
*                         default is the Bandwidth test..
*        -n num_times   : number of times to repeat the send-recv
*                         message passing sequence.
*        -b buff_size   : message buffer size in bytes.
*        -s sync|async  : message passing style; sync is the default.
*        -h             : print usage line.
*        -a             : print out info line for each MPI process.
*
*     Algorithm:
*
*       if (test_type == latency)
*          msg_size = 40 / BUFF_KIND
*       elseif (test_type == bandwidth)
*          msg_size = buff_size / BUFF_KIND
*       endif
*
*       Synchronous:
*
*       t0 = MPI_Wtime()
*       do i = 1, num_times
*          if (master) then
*             call MPI_Send (buf, msg_size, ...)
*             call MPI_Recv (buf, msg_size, ...)
*          else
*             call MPI_Recv (buf, msg_size, ...)
*             call MPI_Send (buf, msg_size, ...)
*          endif
*       enddo
*       t1 = MPI_Wtime()
*
*       Asynchronous:
*
*       t0 = MPI_Wtime()
*       do i = 1, num_times
*          if (master) then
*             call MPI_Irecv (buf, msg_size, ...)
*             call MPI_Send (buf, msg_size, ...)
*             call MPI_Wait (...)
*          else
*             call MPI_Irecv (buf, msg_size, ...)
*             call MPI_Send (buf, msg_size, ...)
*             call MPI_Wait (...)
*          endif
*       enddo
*       t1 = MPI_Wtime()
*
*       if (test_type == latency)
*          latency = 1.0e+6 * (t1 - t0) / (2 * num_times)
*       elseif (test_type == bandwidth)
*          total_bytes = BUFF_KIND * msg_size * num_times * 2
*          bandwidth = (1.0e-6 * total_bytes) / (t1 - t0)
*       endif
*
*     Revision History:
*
*     12-Apr-02 -- SAF; converted to Fortran 90/95.
*     13-Jun-02 -- SAF; add header documentation.
*     13-Jun-02 -- SAF; add abort_flag support in getargs() 
*                       interface.
*     24-Jun-02 -- SAF; replaced num_transfers with num_times.
*     11-Jul-02 -- SAF; added support for the sync/async message
*                       passing styles.
*     12-Jul-02 -- SAF; add compile-time selectable data validation
*                       code.
*     23-Dec-03 -- SAF; add support for the global mpi_data_module
*                       and flush_unit_module as well as the local
*                       lbw_definitions_module.
*     23-Dec-03 -- SAF; add support for compile time selection of
*                       real * 4 or real * 8 message passing buffers.
*
************************************************************************

      use mpi_data_module
      use flush_unit_module
      use lbw_definitions_module

      implicit none

      character (len = 4), parameter :: VERSION = "2.1a"

      integer (kind = 4), parameter :: SRC_RANK  = 0
      integer (kind = 4), parameter :: DEST_RANK = 1

      real (kind = 8) :: t0, t1, trans_time, bandwidth, latency
      integer (kind = 4) :: i, status(MPI_STATUS_SIZE),
     &                      msg_size, num_times
      real (kind = 4) total_bytes

      integer (kind = 4) :: abort_flag, announce_flag, style_flag, test_type

      real (kind = BUFF_KIND), parameter :: magicval= 1.23456
      integer (kind = 4) :: buff_size
      real (kind = BUFF_KIND), dimension (:), allocatable :: in_buff, out_buff
      integer (kind = 4) :: alloc_status

      integer(kind = 4) :: mpi_req

#ifdef VALIDATE
      integer :: j
#endif

*     Begin lbw.

*     Start up MPI.

      call MPI_Begin

*     Write out program header.

      if (my_rank .eq. MASTER_RANK) then
         write (6, 99) VERSION
         write (6, 111) BUFF_KIND
      endif

*     Check to see if we have two processes active. If not, abort.

      if (num_procs .ne. 2) then
         if (my_rank .eq. MASTER_RANK) then 
            write (6, *) 'lbw: must be run using just two MPI processes',
     &                   ', num_procs = ', num_procs
            call flush_unit (6)
         endif
         call MPI_End 
         call exit (1)
      endif

*     If this is the master, interrogate the command line options.

      if (my_rank .eq. MASTER_RANK) then
         call get_command_line_options (num_times, test_type, 
     &                                  buff_size, announce_flag, 
     &                                  abort_flag, style_flag)
      endif

*     Broadcast the abort flag and if some error occurred in the 
*     processing of the command line options, then terminate the 
*     program.

      call MPI_Bcast (abort_flag, 1, MPI_INTEGER, 0, 
     &                MPI_COMM_WORLD, mpi_ierr)
      if (abort_flag .eq. YES) then
         if (my_rank .eq. MASTER_RANK) then
             call print_usage
         endif
         call MPI_End
         call exit (0)
      endif

*     Broadcast arguments to other process.

      call MPI_Bcast (num_times,     1, MPI_INTEGER, 0, 
     &                MPI_COMM_WORLD, mpi_ierr)
      call MPI_Bcast (test_type,     1, MPI_INTEGER, 0, 
     &                MPI_COMM_WORLD, mpi_ierr)
      call MPI_Bcast (buff_size,     1, MPI_INTEGER, 0, 
     &                MPI_COMM_WORLD, mpi_ierr)
      call MPI_Bcast (announce_flag, 1, MPI_INTEGER, 0,
     &                MPI_COMM_WORLD, mpi_ierr)
      call MPI_Bcast (style_flag,    1, MPI_INTEGER, 0,
     &                MPI_COMM_WORLD, mpi_ierr)

*     If the announce flag is set, then call announce().

      if (announce_flag .eq. YES) then
         call announce
      endif

*     Determine message size from test type and precision.

      msg_size = buff_size / BUFF_KIND

*     Allocate and initialize the message buffer.

      allocate (in_buff(msg_size), stat = alloc_status)
      if (alloc_status .ne. 0) then
         write (6, *) "lbw: input message buffer allocation error", 
     &                ", alloc_status = ", alloc_status
         call exit (1)
      endif

      allocate (out_buff(msg_size), stat = alloc_status)
      if (alloc_status .ne. 0) then
         write (6, *) "lbw: output message buffer allocation error", 
     &                ", alloc_status = ", alloc_status
         call exit (1)
      endif

      do i = 1, msg_size
         in_buff(i)  = magicval
         out_buff(i) = magicval
      end do

*     Synchronize all processors.

      call MPI_Barrier (MPI_COMM_WORLD, mpi_ierr)

*     Perform transfers.

      if (style_flag .eq. STYLE_SYNC) then

         if (my_rank .eq. MASTER_RANK) then
            t0 = MPI_Wtime()
         endif
         do i = 1, num_times
            if (my_rank .eq. SRC_RANK) then
#ifdef VALIDATE
               call val_preset_buff (out_buff, msg_size, SRC_RANK)
#endif
               call MPI_Send (out_buff, msg_size, MPI_FLPT, DEST_RANK, i, 
     &                        MPI_COMM_WORLD, mpi_ierr)
               call MPI_Recv (in_buff, msg_size, MPI_FLPT, DEST_RANK, i, 
     &                        MPI_COMM_WORLD, status, mpi_ierr)
#ifdef VALIDATE
               call val_verify_buff (in_buff, msg_size, DEST_RANK)
#endif
            else
               call MPI_Recv (in_buff, msg_size, MPI_FLPT, SRC_RANK, i, 
     &                        MPI_COMM_WORLD, status, mpi_ierr)
#ifdef VALIDATE
               call val_verify_buff (in_buff, msg_size, SRC_RANK)
               call val_preset_buff (out_buff, msg_size, DEST_RANK)
#endif
               call MPI_Send (out_buff, msg_size, MPI_FLPT, SRC_RANK, i, 
     &                        MPI_COMM_WORLD, mpi_ierr)
            endif

         enddo
         if (my_rank .eq. MASTER_RANK) then
            t1 = MPI_Wtime()
         endif

      elseif (style_flag .eq. STYLE_ASYNC) then

         if (my_rank .eq. MASTER_RANK) then
            t0 = MPI_Wtime()
         endif
         do i = 1, num_times
            if (my_rank .eq. SRC_RANK) then
               call MPI_Irecv (in_buff, msg_size, MPI_FLPT, DEST_RANK, i, 
     &                         MPI_COMM_WORLD, mpi_req, mpi_ierr)
#ifdef VALIDATE
               call val_preset_buff (out_buff, msg_size, SRC_RANK)
#endif
               call MPI_Send (out_buff, msg_size, MPI_FLPT, DEST_RANK, i, 
     &                        MPI_COMM_WORLD, mpi_ierr)
               call MPI_Wait (mpi_req, status, mpi_ierr)
#ifdef VALIDATE
               call val_verify_buff (in_buff, msg_size, DEST_RANK)
#endif
            else
               call MPI_Irecv (in_buff, msg_size, MPI_FLPT, SRC_RANK, i, 
     &                        MPI_COMM_WORLD, mpi_req, mpi_ierr)
#ifdef VALIDATE
               call val_preset_buff (out_buff, msg_size, DEST_RANK)
#endif
               call MPI_Send (out_buff, msg_size, MPI_FLPT, SRC_RANK, i, 
     &                        MPI_COMM_WORLD, mpi_ierr)
               call MPI_Wait (mpi_req, status, mpi_ierr)
#ifdef VALIDATE
               call val_verify_buff (in_buff, msg_size, SRC_RANK)
#endif
            endif
         enddo
         if (my_rank .eq. MASTER_RANK) then
            t1 = MPI_Wtime()
         endif

      endif

      if (my_rank .eq. MASTER_RANK) then

         total_bytes = real (BUFF_KIND * msg_size) * 
     &                 real (num_times * 2)
         trans_time = t1 - t0

         if (style_flag .eq. STYLE_SYNC) then
            write (6, 110) "Synchronous"
         elseif (style_flag .eq. STYLE_ASYNC) then
            write (6, 110) "Asynchronous"
         endif
            
         write (6, 100) trans_time
         write (6, 101) num_times
         write (6, 102) real (num_times) / trans_time
         write (6, 103) BUFF_KIND * msg_size
         write (6, 104) total_bytes / 1.0e+06

         if (test_type .eq. TST_LAT) then
            latency = (1.0e+6 *trans_time) / real(2 * num_times)
            write (6, 105) latency
         endif
         if (test_type .eq. TST_BW) then
            bandwidth = (1.0e-6 * total_bytes) / trans_time
            write (6, 106) bandwidth
         endif

      endif

      call MPI_End

*     End lbw.

  99  format (' -- ****** LBW -- Latency and Bandwidth Test,',
     &        ' version = ', a, ' ****** --')
 100  format (' Total transfer time:           ', f12.3, ' sec.')
 101  format (' Number of transfers:           ', i12)
 102  format (' Transfer rate:                 ', f12.1, ' transfers/sec.')
 103  format (' Message size:                  ', i12,   ' bytes')
 104  format (' Total number of bytes:         ', f12.3, ' 10e+06 bytes')
 105  format (' Latency:                       ', f12.2, ' microsec.')
 106  format (' Bandwidth:                     ', f12.3, ' 10e+06 Bytes/sec')

 110  format (' Message Passing Style: ', a)
 111  format (' Message passing data size: ', i2, ' bytes')

      end program lbw

      subroutine get_command_line_options (num_times, test_type, 
     &                                     buff_size, announce_flag, 
     &                                     abort_flag, style_flag)
************************************************************************
*  
*     Name: get_command_line_options.
*
*     Purpose: get command line options.
*
*     Revision History:
*
*     13-Jun-02 -- SAF; add header documentation.
*     13-Jun-02 -- SAF; add support for announce_flag and abort_flag.
*     17-Jun-02 -- SAF; added code to better handle bad or missing
*                       arguments for the '-n' and -L/B options
*     11-Jul-02 -- SAF; added support for the sync/async message
*                       passing styles.
*     24-Dec-03 -- SAF; use mpi_data_module and lbw_definitions_module.
*     25-Dec-03 -- SAF; rename as get_commad_line_options()
*     25-Dec-03 -- SAF; make bandwidth test the default test type.
*
************************************************************************

      use mpi_data_module
      use lbw_definitions_module

      implicit none

      integer (kind = 4) :: num_times, test_type, buff_size

      character (len = 120) :: arg
      integer (kind = 4) num_args, iargc, inx_arg
      integer (kind = 4) :: announce_flag, abort_flag, style_flag
      integer (kind = 4) :: num_times_found, test_type_found, 
     &                      buff_size_found

*     Begin get_command_line_options.

*     Scan through the arguments.

      num_times_found = NO
      buff_size_found = NO
      abort_flag      = NO
      announce_flag   = NO
      style_flag      = STYLE_SYNC
      buff_size       = DEFAULT_BANDWIDTH_BUFFER_SIZE
      test_type       = DEFAULT_TEST_TYPE
      num_args        = iargc()
      do inx_arg = 1, num_args

         call getarg (inx_arg, arg)

         if (arg .eq. '-h') then
            abort_flag = YES
            return
         endif

         if (arg .eq. '-a') then
            announce_flag = YES
         endif

         if (arg .eq. '-B') then
            test_type = TST_BW
         endif

         if (arg .eq. '-L') then
            test_type = TST_LAT
         endif

         if (arg .eq. '-n') then

            num_times_found = YES
            if ((inx_arg + 1) .le. num_args) then

               call getarg (inx_arg + 1, arg)

               read (arg, *, err = 110) num_times
               go to 111
 110           write (6, *) "lbw: undecipherable argument for the -n option"
               abort_flag = YES
               return
 111           continue

            else

               write (6, *) "lbw: missing argument for the -n option"
               abort_flag = YES
               return

            endif

         endif

         if (arg .eq. '-b') then

            buff_size_found = YES
            if ((inx_arg + 1) .le. num_args) then

               call getarg (inx_arg + 1, arg)

               read (arg, *, err = 120) buff_size
               go to 121
 120           write (6, *) "lbw: undecipherable argument for the -b option"
               abort_flag = YES
               return
 121           continue

            else

               write (6, *) "lbw: missing argument for the -b option"
               abort_flag = YES
               return

            endif

         endif

         if (arg .eq. '-s') then

            if ((inx_arg + 1) .le. num_args) then

               call getarg (inx_arg + 1, arg)

               if (arg .eq. "sync") then
                  style_flag = STYLE_SYNC
               elseif (arg .eq. "async") then
                  style_flag = STYLE_ASYNC
               else
                  write (6, *) "lbw: unknown argument for the -s option"
                  abort_flag = YES
               endif

            else

               write (6, *) "lbw: missing argument for the -s option"
               abort_flag = YES
               return

            endif

         endif

      enddo

*     Various sanity checks.

      if (num_times_found .eq. NO) then
         write (6, *) "lbw: missing required argument -n"
         abort_flag = YES
      elseif (num_times .le. 0) then
         write (6, *) "lbw: illegal value for argument -n option"
         abort_flag = YES
      endif

      if (buff_size_found .eq. NO) then
         if (test_type .eq. TST_BW) then 
            buff_size = DEFAULT_BANDWIDTH_BUFFER_SIZE
         endif
         if (test_type .eq. TST_LAT) then 
            buff_size = DEFAULT_LATENCY_BUFFER_SIZE
         endif
      endif

      if (mod(num_times, 1) .eq. 1) then
         num_times = num_times + 1
      endif

      if (buff_size_found .eq. YES) then
         if (buff_size .le. 0) then
            write (6, *) "lbw: illegal value for argument -b option"
            abort_flag = YES
         endif
      endif

*     End get_command_line_options.

      return
      end subroutine get_command_line_options

      subroutine print_usage
************************************************************************
*  
*     Name: print_usage
*
*     Purpose: print usage message to <stdout>.
*
*     Revision History:
*
*     14-Apr-02 -- SAF; add in-line documentation.
*     11-Jul-02 -- SAF; added support for the sync/async message
*                       passing styles.
*     25-Dec-03 -- SAF; add support for optional test type.
*
************************************************************************

*     Begin print_usage.

      write (6, *) 
     & "Usage: lbw -n num_times [-L|-B] [-b buff_size] [-s sync|async] [-h] [-a]"

*     End print_usage.

      return
      end subroutine print_usage

#ifdef VALIDATE
      subroutine val_preset_buff (buff, buff_size, token)
************************************************************************
*  
*     Name: val_preset_buff.
*
*     Purpose: preset buffer contents.
*
*     Revision History:
*
*     12-Jul-02 -- SAF; initial implementation of val_preset_buff().
*
************************************************************************

      implicit none

      real (kind = BUFF_KIND), dimension (*) :: buff
      integer (kind = 4) :: buff_size, token

      integer (kind = 4) :: inx

*     Begin val_preset_buff.

      do inx = 1, buff_size
         buff(inx) = real (token)
      enddo

*     End val_preset_buff.

      return
      end subroutine val_preset_buff

      subroutine val_verify_buff (buff, buff_size, token)
************************************************************************
*  
*     Name: val_verify_buff.
*
*     Purpose: verify buffer contents.
*
*     Revision History:
*
*     12-Jul-02 -- SAF; initial implementation of val_verify_buff().
*
************************************************************************

      implicit none

      real (kind = BUFF_KIND), dimension (*) :: buff
      integer (kind = 4) :: buff_size, token

      integer (kind = 4) :: inx

*     Begin val_verify_buff.

      do inx = 1, buff_size
         if (buff(inx) .ne. real (token)) then
            write (6, 600) inx, buff(inx), real (token)
            call exit (1)
         endif
      enddo

 600  format ("val_verify_buff: bad data compare, inx = ", i10, 
     &        ", buff(inx) = ", e20.10, ", real(token) = ", e20.10)

*     End val_verify_buff.

      return
      end subroutine val_verify_buff
#endif
