**#define MPI_DEBUG
**#define VALIDATE

      include "flush_unit_module.inc"
      include "mpi_data_module.inc"

      module sts_definitions_module

************************************************************************
*  
*     Name: sts_definitions_module
*
*     Usage: sts definitions and global storage.
*
*     use sts_definitions_module
*
*     Revision History:
*
*     20-Dec-03 -- SAF; initial construction of sts_definitions_module.
*       
************************************************************************

      implicit none
      save 

      integer (kind = 4), parameter :: NO = 0, YES = 1
      integer (kind = 4), parameter :: OFF = 0, ON = 1
      integer (kind = 4), parameter :: SRC = 1, DST = 2
      integer (kind = 4), parameter :: MAXIMUM_MESSAGE_LENGTH = 32768
      integer (kind = 4), parameter :: NUM_MSGS_DEFAULT  = 6
      integer (kind = 4), parameter :: NUM_TIMES_DEFAULT = 10
      integer (kind = 4), parameter :: STYLE_SYNC = 0, STYLE_ASYNC = 1

      end module sts_definitions_module
   
      module sts_command_line_options_module

************************************************************************
*  
*     Name: sts_command_line_options_module
*
*     Usage: command line option data definitions and global storage.
*
*     use sts_command_line_options_module
*
*     Revision History:
*
*     20-Dec-02 -- SAF; initial construction of command_line_options module.
*     20-Dec-03 -- SAF; prepend sts_ to module name.
*       
************************************************************************

      implicit none
      save 

      integer (kind = 4) :: num_msgs, num_times, verbose_flag,
     &                      announce_flag, abort_flag, 
     &                      single_flag, double_flag,
     &                      style_flag, rand_seed

      end module sts_command_line_options_module
   
      program sts

************************************************************************
*  
*     Name: sts
*
*     Purpose: SomeToSome MPI benchmark.
*
*     Usage:
*
*     sts -n num_times [-m num_msgs] [-S|-D] [-s sync|async] [-v] [-h] [-a]
*
*        -n num_times    : number of times to repeat the SomeToSome 
*                          operation.
*        -m num_msgs     : number of messages sent by each MPI process.
*        -S|-D           : single (-S) or double (-D) message passing
*                          styles; -S is the default.
*        -s sync|async   : message passing style; sync is the default.
*        -v              : verbose flag for listing of the intermediate 
*                          interconnection map and buffer sizes.
*        -r rand_seed    : random number generator seed; rand_seed > 0.
*        -h              : print usage line.
*        -a              : print out info line for each MPI process.
*
*     Revision History:
*
*     10-Mar-02 -- SAF; initial development of sts().
*     17-Jun-02 -- SAF; double check and fine tune the normal 
*                       and verbose output.
*     11-Jul-02 -- SAF; added support for the sync/async message
*                       passing styles.
*     20-Dec-02 -- SAF; added user specifiable random number seed.
*     20-Dec-02 -- SAF; use command_line_options module.
*     19-Dec-03 -- SAF; up version to 1.1a for the new my_ranf() 
*                       capability.
*     20-Dec-03 -- SAF; use sts_definitions_module.
*     22-Dec-03 -- SAF; use mpi_data_module.
*     
************************************************************************

      use sts_command_line_options_module
      use sts_definitions_module
      use mpi_data_module

      implicit none

      character (len = 4), parameter :: VERSION = "2.1a"

      integer (kind = 4) :: inx, jnx
      integer (kind = 4) :: istat

      integer (kind = 4), allocatable, dimension(:,:) :: iface_list
      integer (kind = 4), allocatable, dimension(:)   :: iface_cnt

      integer (kind = 4) :: send_buffer_size, recv_buffer_size
      real (kind = BUFF_KIND), allocatable, dimension(:) :: send_buffer
      real (kind = BUFF_KIND), allocatable, dimension(:) :: recv_buffer
      integer (kind = 4) :: ave_send_buffer_size, ave_recv_buffer_size

      integer (kind = 4) :: num_ifaces
      real (kind = 8) :: time0, time1, run_time1, run_time2

      integer (kind = 8) num_bytes_transferred1, num_bytes_transferred2
      real (kind = 8) :: total_bandwidth, average_bandwidth_per_iface,
     &                   average_bandwidth_per_proc

*     Begin sts.

*     Initialize the MPI library.

      call MPI_Begin 

*     Write out program header and validation and data size messages.

      if (my_rank .eq. MASTER_RANK) then
         write (6, 600) VERSION
#ifdef VALIDATE
         write (6, 640) 'enabled'
#else
         write (6, 640) 'disabled'
#endif
         write (6, 641) BUFF_KIND
      endif

*     Get the command line options.

      if (my_rank .eq. MASTER_RANK) then
         call get_command_line_options 
      endif

*     Broadcast the abort flag and if some error occurred in the
*     processing of the command line options, then terminate the
*     program.

      call MPI_Bcast (abort_flag, 1, MPI_INTEGER, 0,
     &                MPI_COMM_WORLD, mpi_ierr)
      if (abort_flag .eq. ON) then

         if (my_rank .eq. MASTER_RANK) then
            call print_usage
         endif

         call MPI_End
         call exit (1)

      endif

*     Distribute the command line options to all the MPI processes.

      call MPI_Bcast (num_msgs,      1, MPI_INTEGER, 0, 
     &                MPI_COMM_WORLD, mpi_ierr)
      call MPI_Bcast (num_times,     1, MPI_INTEGER, 0, 
     &                MPI_COMM_WORLD, mpi_ierr)
      call MPI_Bcast (verbose_flag,  1, MPI_INTEGER, 0, 
     &                MPI_COMM_WORLD, mpi_ierr)
      call MPI_Bcast (single_flag,   1, MPI_INTEGER, 0, 
     &                MPI_COMM_WORLD, mpi_ierr)
      call MPI_Bcast (double_flag,   1, MPI_INTEGER, 0, 
     &                MPI_COMM_WORLD, mpi_ierr)
      call MPI_Bcast (announce_flag, 1, MPI_INTEGER, 0,
     &                MPI_COMM_WORLD, mpi_ierr)
      call MPI_Bcast (abort_flag,    1, MPI_INTEGER, 0,
     &                MPI_COMM_WORLD, mpi_ierr)
      call MPI_Bcast (style_flag,    1, MPI_INTEGER, 0,
     &                MPI_COMM_WORLD, mpi_ierr)

*     If the announce flag and is set, then call announce().

      if (announce_flag .eq. ON) then
         call announce
      endif

*     Allocate memory for the interface list and message size counts list.

      if (single_flag .eq. ON) then
         num_ifaces = num_msgs * num_procs
      endif
      if (double_flag .eq. ON) then
         num_ifaces = 2 * num_msgs * num_procs
      endif
      
      allocate (iface_list(2, num_ifaces), stat = istat)
      if (istat .ne. 0) then
         write (6, 695) "iface_list", my_rank, istat
         call MPI_End
         call exit (1)
      endif

      allocate (iface_cnt(num_ifaces), stat = istat)
      if (istat .ne. 0) then
         write (6, 695) "iface_cnt", my_rank, istat
         call MPI_End
         call exit (1)
      endif

*     Randomly assign the interface message pairs and message lengths.

      if (my_rank .eq. MASTER_RANK) then
         call setup_ifaces (num_ifaces, iface_list, iface_cnt)
      endif

*     Distribute the interface list and message counts list.

      call dist_ifaces (num_ifaces, iface_list, iface_cnt)

*     Determine the per process message buffer sizes and allocate
*     memory for the buffers.

      call get_buffer_sizes (num_ifaces, 
     &                       iface_list, iface_cnt,
     &                       send_buffer_size, 
     &                       recv_buffer_size,
     &                       ave_send_buffer_size,
     &                       ave_recv_buffer_size)

      allocate (send_buffer(send_buffer_size), stat = istat)
      if (istat .ne. 0) then
         write (6, 695) "send_buffer", my_rank, istat
         call MPI_End
         call exit (1)
      endif

      allocate (recv_buffer(recv_buffer_size), stat = istat)
      if (istat .ne. 0) then
         write (6, 695) "recv_buffer", my_rank, istat
         call MPI_End
         call exit (1)
      endif

*     Main loop over the SomeToSome operation.

      if (my_rank .eq. MASTER_RANK) then
         time0 = MPI_Wtime()
      endif
      do inx = 1, num_times

         call MPI_Sometosome (num_ifaces, 
     &                        iface_list, iface_cnt, 
     &                        send_buffer, send_buffer_size, 
     &                        recv_buffer, recv_buffer_size)

      enddo
      if (my_rank .eq. MASTER_RANK) then
         time1 = MPI_Wtime()
      endif

*     Compute and print out run-time statistics.

      if (my_rank .eq. MASTER_RANK) then
         num_bytes_transferred1 = 0
         do inx = 1, num_ifaces
            num_bytes_transferred1 = num_bytes_transferred1 + iface_cnt(inx)
         enddo
         num_bytes_transferred1 = BUFF_KIND * num_bytes_transferred1 * num_times
         run_time1 = time1 - time0
         total_bandwidth = real(num_bytes_transferred1) / run_time1

         run_time2 = (time1 - time0) / real (num_times)
         num_bytes_transferred2 = num_bytes_transferred1 / num_times
         Average_bandwidth_per_proc  = (real (num_bytes_transferred2) / real (num_procs))  / real (run_time2)
         average_bandwidth_per_iface = (real (num_bytes_transferred2) / real (num_ifaces)) / real (run_time2)

         if (single_flag .eq. ON) then
            if (style_flag .eq. STYLE_SYNC) then 
               write (6, 601) 'Synchronous', 'Single Sided'
            else if (style_flag .eq. STYLE_ASYNC) then 
               write (6, 601) 'Asynchronous', 'Single Sided'
            endif
         endif
         if (double_flag .eq. ON) then
            if (style_flag .eq. STYLE_SYNC) then 
               write (6, 601) 'Synchronous', 'Double Sided'
            else if (style_flag .eq. STYLE_ASYNC) then 
               write (6, 601) 'Asynchronous', 'Double Sided'
            endif
         endif
         write (6, 609)
         write (6, 610) num_procs
         write (6, 611) num_msgs
         write (6, 612) num_times
         write (6, 630) (BUFF_KIND * (ave_send_buffer_size + ave_recv_buffer_size)) / 2
         if (single_flag .eq. ON) then
            write (6, 631) (BUFF_KIND * (ave_send_buffer_size + ave_recv_buffer_size)) / 
     &                     (2 * num_msgs)
         endif
         if (double_flag .eq. ON) then
            write (6, 631) (BUFF_KIND * (ave_send_buffer_size + ave_recv_buffer_size)) / 
     &                     (4 * num_msgs)
         endif
         write (6, 625) real (num_bytes_transferred1) * 1.0 e-06 
         write (6, 613) run_time1
         write (6, 622) total_bandwidth * 1.0 e-06
         write (6, 619) 
         write (6, 621) (real (num_bytes_transferred2) / real (num_procs)) * 1.0 e-06 
         write (6, 620) run_time2 * 1.0e+03
         write (6, 623) average_bandwidth_per_proc * 1.0e-06
*         write (6, 624) average_bandwidth_per_iface * 1.0e-06
      endif

*     Terminate the MPI library and exit.

      call MPI_End
      call exit (0)

*     End sts.

 600  format (' -- ****** SomeToSome MPI Benchmark, version ', a, 
     &        ' ****** --')
 601  format (' Message passing style: ', a, ', ', a)
 609  format (' Global Statistics:')
 610  format (' Number of procs:                   ', i8)
 611  format (' Number of messages / proc:         ', i8)
 612  format (' Number of passes:                  ', i8)
 630  format (' Average buffer size / proc:    ', i12, ' Bytes')
 631  format (' Average message size / proc:   ', i12, ' Bytes')
 613  format (' Total time:                        ', f8.3, ' seconds')
 619  format (' Per Pass Statistics:')
 620  format (' Total time:                        ', f8.3, ' milliseconds.')
 621  format (' Total bytes / proc:            ', f12.3,
     &        ' 10e+06 Bytes')
 622  format (' Bandwidth total:               ', f12.3,
     &        ' 10e+06 Bytes/sec')
 623  format (' Bandwidth / proc:                  ', f8.3,
     &        ' 10e+06 Bytes/sec/proc')
* 624  format (' Bandwidth / msg:                   ', f8.3,
*     &        ' 10e+06 Bytes/sec/msg')
 625  format (' Total bytes transferred:       ', f12.3,
     &        ' 10e+06 Bytes')
 640  format (' Message passing data validation ', a, '.')
 641  format (' Message passing data size: ', i2, ' Bytes.')

 695  format (' Error allocating ', a, ', rank = ', i6, ', istat = ', i6)

      end program sts

      subroutine get_command_line_options 
************************************************************************
*  
*     Name: get_command_line_options
*
*     Purpose: get command line options.
*
*     Revision History:
*
*     10-Mar-02 -- SAF; initial development of get_command_line_options().
*     13-Jun-02 -- SAF; add support for announce_flag and abort_flag.
*     11-Jul-02 -- SAF; added support for the sync/async message
*                       passing styles.
*     20-Dec-02 -- SAF; add random number seed.
*     20-Dec-02 -- SAF; use command_line_options module.
*     15-Dec-03 -- SAF; add test for positive definite rand_seed; modify
*                       default for my_ranf() support.
*     20-Dec-03 -- SAF; use sts_definitions_module.
*     26-Dec-03 -- SAF; improve sanity checks.
*
************************************************************************

      use sts_command_line_options_module
      use sts_definitions_module

      implicit none

      character (len = 120) :: arg
      integer (kind = 4) :: num_args, iargc, inx_arg
      integer (kind = 4) :: num_times_found      

*     Begin get_command_line_options.

*     Initialize command line options.

      abort_flag      = OFF
      num_msgs        = NUM_MSGS_DEFAULT
      num_times       = NUM_TIMES_DEFAULT
      num_times_found = NO
      verbose_flag    = OFF
      announce_flag   = OFF
      single_flag     = ON
      double_flag     = OFF
      style_flag      = STYLE_SYNC
      rand_seed       = 123456

*     Scan through the arguments.

      num_args = iargc()
      do inx_arg = 1, num_args

         call getarg (inx_arg, arg)

         if (arg .eq. '-h') then
            abort_flag = ON
            return
         endif

         if (arg .eq. '-a') then
            announce_flag = ON
         endif

         if (arg .eq. '-S') then
            single_flag = ON
            double_flag = OFF
         endif

         if (arg .eq. '-D') then
            single_flag = OFF
            double_flag = ON
         endif

         if (arg .eq. '-m') then

            if (inx_arg + 1 .le. num_args) then

               call getarg (inx_arg + 1, arg)

               read (arg, *, err = 105) num_msgs
               go to 106
 105           write (6, *) "sts: undecipherable argument for the -m option"
               abort_flag = ON
               return
 106           continue

            else

               write (6, *) "sts: missing argument for the -m option"
               abort_flag = ON
               return

            endif

         endif

         if (arg .eq. '-n') then

            num_times_found = YES
            if (inx_arg + 1 .le. num_args) then

               call getarg (inx_arg + 1, arg)

               read (arg, *, err = 110) num_times
               go to 111
 110           write (6, *) "sts: undecipherable argument for the -n option"
               abort_flag = ON
               return
 111           continue

            else

               write (6, *) "sts: missing argument for the -n option"
               abort_flag = ON
               return

            endif

         endif

         if (arg .eq. '-s') then

            if ((inx_arg + 1) .le. num_args) then

               call getarg (inx_arg + 1, arg)
               if (arg .eq. "sync") then
                  style_flag = STYLE_SYNC
               elseif (arg .eq. "async") then
                  style_flag = STYLE_ASYNC
               else
                  write (6, *) "sts: unknown argument for the -s option"
                  abort_flag = ON
               endif

            else

               write (6, *) "sts: missing argument for the -s option"
               abort_flag = ON
               return

            endif

         endif

         if (arg .eq. '-r') then

            if ((inx_arg + 1) .le. num_args) then

               call getarg (inx_arg + 1, arg)
               read (arg, *) rand_seed
               if (rand_seed .le. 0) then
                  write (6, *) "sts: rand_seed = ", rand_seed
                  write (6, *) "sts: illegal random seed, rand_seed > 0"
                  abort_flag = ON
                  return
               endif

            else

               write (6, *) "sts: missing argument for the -r option"
               abort_flag = ON
               return

            endif

         endif

         if (arg .eq. '-v') then
            verbose_flag = ON
         endif

      enddo

*     Sanity checks.

      if (num_times_found .eq. NO) then
         write (6, *) "sts: missing required argument -n"
         abort_flag = ON
      endif
      if (num_times .le. 0) then
         write (6, *) "sts: illegal value for argument -n option"
         abort_flag = ON
      endif

      if (num_msgs .le. 0) then
         write (6, *) "sts: illegal value for argument -m option"
         abort_flag = ON
      endif

*     End get_command_line_options.

      return
      end subroutine get_command_line_options 

      subroutine print_usage
************************************************************************
*  
*     Name: print_usage
*
*     Purpose: print usage line.
*
*     Revision History:
*
*     10-Mar-02 -- SAF; initial development of print_usage().
*     13-Jun-02 -- SAF; add support for announce_flag.
*     16-Jun-02 -- SAF; add support for single/double message passing
*                       flags.
*     11-Jul-02 -- SAF; added support for the sync/async message
*                       passing styles.
*     20-Dec-02 -- SAF; added support for the random number seed.
*
************************************************************************

*     Begin print_usage.

      write (6, '("Usage: sts -n num_times [-m num_msgs] ",
     &            "[-r rand_seed] [-S|-D] [-s sync|async] [-v] [-h] [-a]")')

*     End print_usage.

      return
      end subroutine print_usage

      subroutine setup_ifaces (num_ifaces, iface_list, iface_cnt)
************************************************************************
*  
*     Name: setup_ifaces
*
*     Purpose: randomly assign message pairs and message length counts.
*
*     Revision History:
*
*     10-Mar-02 -- SAF; initial development of setup_ifaces().
*     16-Jun-02 -- SAF; added support for double sided communications.
*     20-Dec-02 -- SAF; add support for rand number seed.
*     20-Dec-02 -- SAF; use command_line_options module.
*     26-Dec-02 -- SAF; add support for platform specific random
*                       number generators.
*     15-Dec-03 -- SAF; replace platform-specific random number routines
*                       with a (hopefully) portable my_ranf() version.
*     20-Dec-03 -- SAF; use sts_definitions_module.
*     22-Dec-03 -- SAF; use mpi_data_module.
*
************************************************************************

      use sts_command_line_options_module
      use sts_definitions_module
      use mpi_data_module

      implicit none

      integer (kind = 4) :: num_ifaces
      integer (kind = 4), dimension(2, num_ifaces) :: iface_list
      integer (kind = 4), dimension(num_ifaces)    :: iface_cnt

      integer (kind = 4) :: rank1, rank2
      integer (kind = 4) :: inx, jnx, knx

      integer (kind = 4) :: msg_buffer_size_sum, ave_msg_buffer_size
      integer (kind = 4) :: itemp

      real (kind = 4) :: my_ranf, rtemp

*     Begin setup_ifaces.

*     Set the random number generator seed.

       if (rand_seed .gt. 0) then
*          write (6, *) "rand_seed = ", rand_seed
          rtemp = my_ranf (-rand_seed)
       endif

*     If verbose flag set, then print out some headers.

      if (verbose_flag .eq. ON) then
         write (6, 600)
         write (6, 601)
         write (6, 602)
      endif

*     Loop through the interfaces assigning process pairs and message
*     length counts.

      jnx = 0
      if (single_flag .eq. ON) then
         knx = num_ifaces
      endif
      if (double_flag .eq. ON) then
         knx = num_ifaces / 2
      endif

      msg_buffer_size_sum = 0
      do inx = 1, knx

 100     continue

         rank1 = num_procs * my_ranf (0)
         rank2 = num_procs * my_ranf (0)

         if (rank1 .eq. rank2) then
            go to 100            ! disallow case where rank1 = rank2
         endif

         jnx = jnx + 1
         iface_list(SRC, jnx) = rank1
         iface_list(DST, jnx) = rank2

 101     continue
            iface_cnt(jnx) = MAXIMUM_MESSAGE_LENGTH * my_ranf(0)
     
            if (iface_cnt(jnx) .eq. 0) then 
               go to 101         ! disallow zero length messages.
            endif

         msg_buffer_size_sum = msg_buffer_size_sum + iface_cnt(jnx)

         if (verbose_flag .eq. ON) then
            write (6, 603) jnx, iface_list(SRC, jnx),
     &                     iface_list(DST, jnx), BUFF_KIND * iface_cnt(jnx)
         endif

         if (double_flag .eq. ON) then
            jnx = jnx + 1
            iface_list(SRC, jnx) = rank2
            iface_list(DST, jnx) = rank1

 102        continue

               iface_cnt(jnx) = MAXIMUM_MESSAGE_LENGTH * my_ranf(0)
               if (iface_cnt(jnx) .eq. 0) then 
                  go to 102         ! disallow zero length messages.
               endif

            msg_buffer_size_sum = msg_buffer_size_sum + iface_cnt(jnx)

            if (verbose_flag .eq. ON) then
               write (6, 603) jnx, iface_list(SRC, jnx),
     &                        iface_list(DST, jnx), BUFF_KIND * iface_cnt(jnx)
            endif
         endif

      enddo

      if (verbose_flag .eq. ON) then
         write (6, 604)
         write (6, 605) (BUFF_KIND * msg_buffer_size_sum) / num_ifaces
         write (6, '("")')  ! blank line
      endif

*     End setup_ifaces.

 600  format (/, "        Interface Map")
 601  format (" Iface   Src   Dest    Length")
 602  format ("======= ===== ====== ==========")
 603  format (i5, 2x, i5, 1x, i5, 2x, i10)
 604  format ("                     ==========")
 605  format ("Ave:                ", i10)

      return
      end subroutine setup_ifaces

      subroutine dist_ifaces (num_ifaces, iface_list, iface_cnt)
************************************************************************
*  
*     Name: dist_ifaces
*
*     Purpose: distribute the interface data to the other MPI processes.
*
*     Revision History:
*
*     10-Mar-02 -- SAF; initial development of dist_ifaces().
*     22-Dec-03 -- SAF; use mpi_data_module.
*
************************************************************************

      use mpi_data_module

      implicit none

      integer (kind = 4) :: num_ifaces
      integer (kind = 4), dimension(2, num_ifaces) :: iface_list
      integer (kind = 4), dimension(num_ifaces)    :: iface_cnt

      integer (kind = 4) :: inx

*     Begin dist_ifaces.

      call MPI_Bcast (iface_list, 2 * num_ifaces, MPI_INTEGER, 
     &                0, MPI_COMM_WORLD, mpi_ierr)
      call MPI_Bcast (iface_cnt, num_ifaces, MPI_INTEGER, 
     &                0, MPI_COMM_WORLD, mpi_ierr)

*     End dist_ifaces.

      return
      end subroutine dist_ifaces

      subroutine MPI_Sometosome (num_ifaces, 
     &                           iface_list, iface_cnt, 
     &                           send_buffer, send_buffer_size,
     &                           recv_buffer, recv_buffer_size)
************************************************************************
*  
*     Name: MPI_Sometosome
*
*     Purpose: execute the sometosome operations using an Irecv, Send
*              Waitall sequence of functions.
*
*     Revision History:
*
*     10-Mar-02 -- SAF; initial development of MPI_Sometosome().
*     11-Jul-02 -- SAF; added support for the sync/async message
*                       passing styles.
*     11-Jul-02 -- SAF; converted to using MPI_FLPT.
*     12-Jul-02 -- SAF; add compile-time selectable data validation
*                       code.
*     20-Dec-03 -- SAF; use sts_command_line_options_module.
*     20-Dec-03 -- SAF; use sts_definitions_module.
*     22-Dec-03 -- SAF; use mpi_data_module and flush_unit_module.
*
************************************************************************

      use sts_command_line_options_module
      use sts_definitions_module
      use mpi_data_module
      use flush_unit_module

      implicit none

      integer (kind = 4) :: num_ifaces, send_buffer_size, recv_buffer_size
      integer (kind = 4) :: src_rank, dest_rank, count
      integer (kind = 4), dimension(2, num_ifaces) :: iface_list
      integer (kind = 4), dimension(num_ifaces)    :: iface_cnt

      real (kind = BUFF_KIND), dimension(send_buffer_size) :: send_buffer
      real (kind = BUFF_KIND), dimension(recv_buffer_size) :: recv_buffer

      integer, dimension(MPI_STATUS_SIZE, num_ifaces) :: status_list
      integer, dimension(num_ifaces) :: req_list
      integer (kind = 4) :: inx, req_cnt, send_buff_ptr, recv_buff_ptr

*     Begin MPI_Sometosome.

      if (style_flag .eq. STYLE_SYNC) then

*        Issue MPI_Send calls to send messages to all receiving MPI
*        processes.

         send_buff_ptr = 1
         recv_buff_ptr = 1
         do inx = 1, num_ifaces
            src_rank  = iface_list(SRC, inx)
            dest_rank = iface_list(DST, inx)
            count     = iface_cnt(inx)
            if (my_rank .eq. src_rank) then
#ifdef MPI_DEBUG
               write (6, 660) src_rank, dest_rank, count, inx, send_buff_ptr
               call flush_unit (6)
#endif
#ifdef VALIDATE
               call val_preset_buff (send_buffer(send_buff_ptr), count, src_rank)
#endif
               call MPI_Send (send_buffer(send_buff_ptr), count, MPI_FLPT, 
     &                        dest_rank, inx, MPI_COMM_WORLD, mpi_ierr)
               send_buff_ptr = send_buff_ptr + count
            endif
            if (my_rank .eq. dest_rank) then
#ifdef MPI_DEBUG
               write (6, 670) src_rank, dest_rank, count, inx, recv_buff_ptr
               call flush_unit (6)
#endif
               call MPI_Recv (recv_buffer(recv_buff_ptr), count, MPI_FLPT, 
     &                        src_rank, inx, MPI_COMM_WORLD, status_list, 
     &                        mpi_ierr)
#ifdef VALIDATE
               call val_verify_buff (recv_buffer(recv_buff_ptr), 
     &                               count, src_rank)
#endif
               recv_buff_ptr = recv_buff_ptr + count
            endif
         enddo

      elseif (style_flag .eq. STYLE_ASYNC) then

*        Issue MPI_Irecv() functions to catch all the incoming data.

         req_cnt = 0
         recv_buff_ptr = 1
         do inx = 1, num_ifaces
            src_rank  = iface_list(SRC, inx)
            dest_rank = iface_list(DST, inx)
            count     = iface_cnt(inx)
            if (my_rank .eq. dest_rank) then
#ifdef MPI_DEBUG
               write (6, 670) src_rank, dest_rank, count, inx, recv_buff_ptr
               call flush_unit (6)
#endif
               req_cnt = req_cnt + 1
               call MPI_Irecv (recv_buffer(recv_buff_ptr), count, 
     &                         MPI_FLPT, src_rank, inx, 
     &                         MPI_COMM_WORLD, req_list(req_cnt), 
     &                         mpi_ierr)
               recv_buff_ptr = recv_buff_ptr + count
            endif
         enddo

*        Issue MPI_Send calls to send messages to all receiving MPI
*        processes.

         send_buff_ptr = 1
         do inx = 1, num_ifaces
            src_rank  = iface_list(SRC, inx)
            dest_rank = iface_list(DST, inx)
            count     = iface_cnt(inx)
            if (my_rank .eq. src_rank) then
#ifdef MPI_DEBUG
               write (6, 660) src_rank, dest_rank, count, inx, send_buff_ptr
               call flush_unit (6)
#endif
#ifdef VALIDATE
               call val_preset_buff (send_buffer(send_buff_ptr), 
     &                               count, src_rank)
#endif
               call MPI_Send (send_buffer(send_buff_ptr), count, MPI_FLPT, 
     &                        dest_rank, inx, MPI_COMM_WORLD, mpi_ierr)
               send_buff_ptr = send_buff_ptr + count
            endif
         enddo

*        Wait for all incoming messages.

         call MPI_Waitall (req_cnt, req_list, status_list, mpi_ierr)

#ifdef VALIDATE
         recv_buff_ptr = 1
         do inx = 1, num_ifaces
            src_rank  = iface_list(SRC, inx)
            dest_rank = iface_list(DST, inx)
            count     = iface_cnt(inx)
            if (my_rank .eq. dest_rank) then
               call val_verify_buff (recv_buffer(recv_buff_ptr), count, src_rank)
               recv_buff_ptr = recv_buff_ptr + count
            endif
         enddo
#endif
      endif

*     End MPI_Sometosome.

#ifdef MPI_DEBUG
 660  format ("sending from ", i3, " to ", i3, ", count = ", i8,
     &        ", tag = ", i3, ", send_buff_ptr = ", i8)
 670  format ("receiving from ", i3, " to ", i3, ", count = ", i8,
     &        ", tag = ", i3, ", recv_buff_ptr = ", i8)
#endif

      return
      end subroutine MPI_Sometosome

      subroutine get_buffer_sizes (num_ifaces, 
     &                            iface_list, iface_cnt,
     &                            send_buffer_size,
     &                            recv_buffer_size,
     &                            ave_send_buffer_size,
     &                            ave_recv_buffer_size)
************************************************************************
*  
*     Name: get_buffer_sizes
*
*     Purpose: gather the buffer sizes from each of the MPI processes.
*
*     Revision History:
*
*     10-Mar-02 -- SAF; initial development of get_buffer_sizes().
*     16-Jun-02 -- SAF; added support ofr double sided communications.
*     20-Dec-03 -- SAF; use sts_definitions_module.
*     22-Dec-03 -- SAF; use mpi_data_module.
*
************************************************************************

      use sts_definitions_module
      use mpi_data_module
      use sts_command_line_options_module
 
      implicit none

      integer (kind = 4) :: num_ifaces
      integer (kind = 4), dimension(2, num_ifaces) :: iface_list
      integer (kind = 4), dimension(num_ifaces)    :: iface_cnt

      integer (kind = 4) :: send_buffer_size, recv_buffer_size
      integer (kind = 4), dimension(num_procs) :: send_buffer_sizes, 
     &                                            recv_buffer_sizes

      integer (kind = 4) :: inx
      integer (kind = 4) :: sum_send_buffer_size, sum_recv_buffer_size
      integer (kind = 4) :: ave_send_buffer_size, ave_recv_buffer_size

*     Begin get_buffer_sizes.

*     Compute the total send and recv buffer sizes for this MPI process.

      send_buffer_size = 0
      do inx = 1, num_ifaces
         if (my_rank .eq. iface_list(SRC, inx)) then
            send_buffer_size = send_buffer_size + iface_cnt(inx)
         endif
      enddo

      recv_buffer_size = 0
      do inx = 1, num_ifaces
         if (my_rank .eq. iface_list(DST, inx)) then
            recv_buffer_size = recv_buffer_size + iface_cnt(inx)
         endif
      enddo

*     Gather up all the buffer sizes and if the verbose output flag is set, 
*     print them out.

      call MPI_Gather (send_buffer_size, 1, MPI_INTEGER,
     &                 send_buffer_sizes, 1, MPI_INTEGER,
     &                 0, MPI_COMM_WORLD, mpi_ierr)

      call MPI_Gather (recv_buffer_size, 1, MPI_INTEGER,
     &                 recv_buffer_sizes, 1, MPI_INTEGER,
     &                 0, MPI_COMM_WORLD, mpi_ierr)

      if ((my_rank .eq. MASTER_RANK) .and. 
     &    (verbose_flag .eq. ON)) then
         write (6, 600)
         write (6, 601)
         write (6, 602)
      endif

      sum_send_buffer_size = 0
      sum_recv_buffer_size = 0
      do inx = 1, num_procs
         if ((my_rank .eq. MASTER_RANK) .and. 
     &       (verbose_flag .eq. ON)) then
            write (6, 603) inx - 1, BUFF_KIND * send_buffer_sizes(inx), 
     &                     BUFF_KIND * recv_buffer_sizes(inx)
         endif
         sum_send_buffer_size = sum_send_buffer_size + send_buffer_sizes(inx)
         sum_recv_buffer_size = sum_recv_buffer_size + recv_buffer_sizes(inx)
      enddo

      if ((my_rank .eq. MASTER_RANK) .and. 
     &    (verbose_flag .eq. ON)) then
         write (6, 610)
      endif
      ave_send_buffer_size = int (real (sum_send_buffer_size) / real(num_procs))
      ave_recv_buffer_size = int (real (sum_recv_buffer_size) / real(num_procs))
      if ((my_rank .eq. MASTER_RANK) .and. 
     &    (verbose_flag .eq. ON)) then
         write (6, 611) BUFF_KIND * ave_send_buffer_size, 
     &                  BUFF_KIND * ave_recv_buffer_size
         write (6, '("")')       ! blank line
      endif

*     End get_buffer_sizes.

 600  format ("     Transfer Buffer Sizes")
 601  format (" Proc      Send         Recv")
 602  format ("====== ============ ============")
 603  format (i4, 3x, i10, 3x, i10)
 610  format ("====== ============ ============")
 611  format (" Ave:  ", i10, 3x, i10)

      return
      end subroutine get_buffer_sizes

#ifdef VALIDATE
      subroutine val_preset_buff (buff, buff_size, token)
************************************************************************
*  
*     Name: val_preset_buff.
*
*     Purpose: preset buffer contents.
*
*     Revision History:
*
*     12-Jul-02 -- SAF; initial implementation of val_preset_buff().
*
************************************************************************

      implicit none

      real (kind = BUFF_KIND), dimension (*) :: buff
      integer (kind = 4) :: buff_size, token

      integer (kind = 4) :: inx

*     Begin val_preset_buff.

      do inx = 1, buff_size
         buff(inx) = real (token)
      enddo

*     End val_preset_buff.

      return
      end subroutine val_preset_buff

      subroutine val_verify_buff (buff, buff_size, token)
************************************************************************
*  
*     Name: val_verify_buff.
*
*     Purpose: verify buffer contents.
*
*     Revision History:
*
*     12-Jul-02 -- SAF; initial implementation of val_verify_buff().
*     22-Dec-03 -- SAF; use flush_unit_module.
*
************************************************************************

      use flush_unit_module

      implicit none

      real (kind = BUFF_KIND), dimension (*) :: buff
      integer (kind = 4) :: buff_size, token

      integer (kind = 4) :: inx

*     Begin val_verify_buff.

      do inx = 1, buff_size
         if (buff(inx) .ne. real (token)) then
            write (6, 600) inx, buff(inx), real (token)
            call flush_unit (6)
            call abort()
            call exit (1)
         endif
      enddo

 600  format ("val_verify_buff: bad data compare, inx = ", i10, 
     &        ", buff(inx) = ", e20.10, ", real(token) = ", e20.10)

*     End val_verify_buff.

      return
      end subroutine val_verify_buff
#endif
      real function my_ranf (iseed)
************************************************************************
*  
*     Name: my_ranf
*
*     Purpose: portable random number generator.
*
*     Revision History:
*
*     15-Dec-03 -- SAF; initial development of my_ranf().
*
************************************************************************

      implicit none

      integer (kind = 4), dimension (97) :: ir
      integer (kind = 4) :: iseed
      integer (kind = 4), parameter :: M = 714025, IA = 1366,  IC = 150889
      real (kind = 4), parameter :: RM = 1.0 / M
      integer (kind = 4) :: iff = 0
      integer (kind = 4) :: idum, inx, iy
      save ir, iy, iff, idum

*     Begin my_ranf.

      if ((iseed .lt. 0) .or. (iff .eq. 0)) then
         if (iseed .lt. 0) then       
            idum = iseed
         else
            idum = -123456
         endif
         iff = 1
         idum = mod (IC - idum, M)
         do inx = 1, 97
            idum = mod ((IA * idum) + IC, M)
            ir(inx) = idum
         enddo
         idum = mod ((IA * idum) + IC, M)
         iy = idum

      endif

      inx = 1 + ((97 * iy) / M)
      if ((inx .lt. 1) .or. (inx .gt. 97)) then
         write (6, *) "my_ranf: index failure, inx = ", inx
         call exit (1)
      endif

      iy = ir(inx)
      my_ranf = real (iy) * RM
      idum = mod ((IA * idum) + IC, M)
      ir (inx) = idum

*     End my_ranf.

      return
      end function my_ranf
